from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Channel:

    @dataclass
    class ModelChannel:
        name: str
        creation_time: str
        num_users: int
        topic: str
        topic_set_by: str
        topic_set_at: str
        modes: str
        bans: list
        ban_exemptions: list
        invite_exceptions: list
        members: list

    DB_CHANNELS: list[ModelChannel] = []


    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self, _object_detail_level: int = 1) -> Union[list[ModelChannel], None]:
        """List channels.

        if you want to have more details increase the level or see the level you want by visiting this page:

        https://www.unrealircd.org/docs/JSON-RPC:Channel#Structure_of_a_channel

        Args:
            _object_detail_level (int, optional): set the detail of the response object, see the Detail level column in Structure of a channel. In this RPC call it defaults to 1 if this parameter is not specified. Defaults to 1.

        Returns:
            ModelChannel: List of ModelChannel, None if nothing see the Error property
        """
        try:
            response = self.Connection.query(method='channel.list', param={'object_detail_level': _object_detail_level})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            channels = response['result']['list']

            for channel in channels:
                self.DB_CHANNELS.append(
                        self.ModelChannel(
                            name=channel['name'] if 'name' in channel else None,
                            creation_time=channel['creation_time'] if 'creation_time' in channel else None,
                            num_users=channel['num_users'] if 'num_users' in channel else 0,
                            topic=channel['topic'] if 'topic' in channel else None,
                            topic_set_by=channel['topic_set_by'] if 'topic_set_by' in channel else None,
                            topic_set_at=channel['topic_set_at'] if 'topic_set_at' in channel else None,
                            modes=channel['modes'] if 'modes' in channel else None,
                            bans=channel['bans'] if 'bans' in channel else [],
                            ban_exemptions=channel['ban_exemptions'] if 'ban_exemptions' in channel else [],
                            invite_exceptions=channel['invite_exceptions'] if 'invite_exceptions' in channel else [],
                            members=channel['members'] if 'members' in channel else []
                        )
                )

            return self.DB_CHANNELS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, channel: str, _object_detail_level: int = 3) -> Union[ModelChannel, None]:
        """Retrieve all details of a single channel. 
        This returns more information than a channel.list call, see the end of Structure of a channel.

        Args:
            channel (str):  the name of the channel
            _object_detail_level (int, optional): set the detail of the response object, see the Detail level column in Structure of a channel. In this RPC call it defaults to 3 if this parameter is not specified. Defaults to 3.

        Returns:
            ModelChannel: The ModelChannel, None if nothing see Error property
        """
        try:
            response = self.Connection.query(method='channel.get', param={'channel': channel, 'object_detail_level': _object_detail_level})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            channel = response['result']['channel']

            objectChannel = self.ModelChannel(
                            name=channel['name'] if 'name' in channel else None,
                            creation_time=channel['creation_time'] if 'creation_time' in channel else None,
                            num_users=channel['num_users'] if 'num_users' in channel else 0,
                            topic=channel['topic'] if 'topic' in channel else None,
                            topic_set_by=channel['topic_set_by'] if 'topic_set_by' in channel else None,
                            topic_set_at=channel['topic_set_at'] if 'topic_set_at' in channel else None,
                            modes=channel['modes'] if 'modes' in channel else None,
                            bans=channel['bans'] if 'bans' in channel else [],
                            ban_exemptions=channel['ban_exemptions'] if 'ban_exemptions' in channel else [],
                            invite_exceptions=channel['invite_exceptions'] if 'invite_exceptions' in channel else [],
                            members=channel['members'] if 'members' in channel else []
                        )

            return objectChannel

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_mode(self, channel: str, modes: str, parameters: str) -> bool:
        """Set and unset modes on a channel.

        Args:
            channel (str): The name of the channel
            modes (str): The mode(s) to change, eg +be
            parameters (str): The parameters, eg some!nice@ban some!nice@invex

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='channel.set_mode', param={"channel": channel,"modes": modes,"parameters": parameters})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_topic(self, channel: str, topic: str, _set_by: str = None, _set_at: str = None) -> bool:
        """Set a topic on a channel.

        Args:
            channel (str): The name of the channel
            topic (str): The new topic
            _set_by (str, optional): who set the topic (but, see important note below). Defaults to None.
            _set_at (str, optional): when the topic was set (timestamp). Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='channel.set_topic', param={"channel": channel, "topic": topic, "set_by": _set_by, "set_at": _set_at})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def kick(self, channel: str, nick: str, reason: str) -> bool:
        """Kick a user from a channel

        Args:
            channel (str): The name of the channel
            nick (str): The user to kick
            reason (str): The kick reason (shown in the channel)

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='channel.kick', param={"channel": channel, "nick": nick, "reason": reason})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')


from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Spamfilter:

    @dataclass
    class ModelSpamfilter:
        type: str
        type_string: str
        set_by: str
        set_at: str
        expire_at: str
        set_at_string: str
        expire_at_string: str
        duration_string: str
        set_at_delta: int
        set_in_config: bool
        name: str
        match_type: str
        ban_action: str
        ban_duration: int
        ban_duration_string: str
        spamfilter_targets: str
        reason: str
        hits: int
        hits_except: int

    DB_SPAMFILTERS: list[ModelSpamfilter] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelSpamfilter], None]:
        """List spamfilters.

        Returns:
            list[ModelSpamfilter]: List of ModelSpamfilter, None if nothing, see Error property if any error
        """
        try:
            response = self.Connection.query(method='spamfilter.list')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            spamfilters = response['result']['list']

            for spamfilter in spamfilters:
                self.DB_SPAMFILTERS.append(
                        self.ModelSpamfilter(
                            type=spamfilter['type'] if 'type' in spamfilter else None,
                            type_string=spamfilter['type_string'] if 'type_string' in spamfilter else None,
                            set_by=spamfilter['set_by'] if 'set_by' in spamfilter else None,
                            set_at=spamfilter['set_at'] if 'set_at' in spamfilter else None,
                            expire_at=spamfilter['expier_at'] if 'expier_at' in spamfilter else None,
                            set_at_string=spamfilter['set_at_string'] if 'set_at_string' in spamfilter else None,
                            expire_at_string=spamfilter['expier_at_string'] if 'expier_at_string' in spamfilter else None,
                            duration_string=spamfilter['duration_string'] if 'duration_string' in spamfilter else None,
                            set_at_delta=spamfilter['set_at_delta'] if 'set_at_delta' in spamfilter else 0,
                            set_in_config=spamfilter['set_in_config'] if 'set_in_config' in spamfilter else False,
                            name=spamfilter['name'] if 'name' in spamfilter else None,
                            match_type=spamfilter['match_type'] if 'match_type' in spamfilter else None,
                            ban_action=spamfilter['ban_action'] if 'ban_action' in spamfilter else None,
                            ban_duration=spamfilter['ban_duration'] if 'ban_duration' in spamfilter else 0,
                            ban_duration_string=spamfilter['ban_duration_string'] if 'ban_duration_string' in spamfilter else None,
                            spamfilter_targets=spamfilter['spamfilter_targets'] if 'spamfilter_targets' in spamfilter else None,
                            reason=spamfilter['reason'] if 'reason' in spamfilter else None,
                            hits=spamfilter['hits'] if 'hits' in spamfilter else 0,
                            hits_except=spamfilter['hits_except'] if 'hits_except' in spamfilter else 0
                            )
                )

            return self.DB_SPAMFILTERS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, name: str, match_type: str, ban_action: str, spamfilter_targets: str) -> Union[ModelSpamfilter, None]:
        """Retrieve all details of a single spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.

        Returns:
            ModelSpamfilter: The Object ModelSpamfilter, None if nothing (see Error property)
        """
        try:
            response = self.Connection.query(method='spamfilter.get', param={"name": name, "match_type": match_type, "ban_action": ban_action, "spamfilter_targets": spamfilter_targets})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            spamfilter = response['result']['tkl']

            objectSpamfilter = self.ModelSpamfilter(
                            type=spamfilter['type'] if 'type' in spamfilter else None,
                            type_string=spamfilter['type_string'] if 'type_string' in spamfilter else None,
                            set_by=spamfilter['set_by'] if 'set_by' in spamfilter else None,
                            set_at=spamfilter['set_at'] if 'set_at' in spamfilter else None,
                            expire_at=spamfilter['expier_at'] if 'expier_at' in spamfilter else None,
                            set_at_string=spamfilter['set_at_string'] if 'set_at_string' in spamfilter else None,
                            expire_at_string=spamfilter['expier_at_string'] if 'expier_at_string' in spamfilter else None,
                            duration_string=spamfilter['duration_string'] if 'duration_string' in spamfilter else None,
                            set_at_delta=spamfilter['set_at_delta'] if 'set_at_delta' in spamfilter else 0,
                            set_in_config=spamfilter['set_in_config'] if 'set_in_config' in spamfilter else False,
                            name=spamfilter['name'] if 'name' in spamfilter else None,
                            match_type=spamfilter['match_type'] if 'match_type' in spamfilter else None,
                            ban_action=spamfilter['ban_action'] if 'ban_action' in spamfilter else None,
                            ban_duration=spamfilter['ban_duration'] if 'ban_duration' in spamfilter else 0,
                            ban_duration_string=spamfilter['ban_duration_string'] if 'ban_duration_string' in spamfilter else None,
                            spamfilter_targets=spamfilter['spamfilter_targets'] if 'spamfilter_targets' in spamfilter else None,
                            reason=spamfilter['reason'] if 'reason' in spamfilter else None,
                            hits=spamfilter['hits'] if 'hits' in spamfilter else 0,
                            hits_except=spamfilter['hits_except'] if 'hits_except' in spamfilter else 0
                        )

            return objectSpamfilter

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add(self, name: str, match_type: str, ban_action: str, ban_duration: int, spamfilter_targets: str, reason: str, _set_by: str = None) -> bool:
        """Add a spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        {"name":"regex123","match_type": "regex","ban_action": "gline","ban_duration": 30,"spamfilter_targets": "cpnNPq","reason": "RPC test"}

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            ban_duration (int): The duration of the ban
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.
            reason (str): The reason of the ban
            _set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='spamfilter.add', 
                param={"name": name, "match_type": match_type, "ban_action": ban_action, "ban_duration": ban_duration, "spamfilter_targets": spamfilter_targets, 'reason': reason, 'set_by': _set_by}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def del_(self,  name: str, match_type: str, ban_action: str, spamfilter_targets: str, _set_by: str = None) -> bool:
        """Delete a spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.
            _set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='spamfilter.del',
                param={"name": name, "match_type": match_type, "ban_action": ban_action, "spamfilter_targets": spamfilter_targets, 'set_by': _set_by}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')


from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Server_ban:

    @dataclass
    class ModelServerBan:
        type: str
        type_string: str
        set_by: str
        set_at: str
        expire_at: str
        set_at_string: str
        expire_at_string: str
        duration_string: str
        set_at_delta: int
        set_in_config: bool
        name: str
        reason: str

    DB_SERVERS_BANS: list[ModelServerBan] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelServerBan], None]:
        """List server bans (LINEs).

        Returns:
            list[ModelServerBan]: List of ModelServerBan, None if nothing see the Error property
        """
        try:
            response = self.Connection.query(method='server_ban.list')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            srvbans = response['result']['list']

            for srvban in srvbans:
                self.DB_SERVERS_BANS.append(
                        self.ModelServerBan(
                            type=srvban['type'] if 'type' in srvban else None,
                            type_string=srvban['type_string'] if 'type_string' in srvban else None,
                            set_by=srvban['set_by'] if 'set_by' in srvban else None,
                            set_at=srvban['set_at'] if 'set_at' in srvban else None,
                            expire_at=srvban['expire_at'] if 'expire_at' in srvban else None,
                            set_at_string=srvban['set_at_string'] if 'set_at_string' in srvban else None,
                            expire_at_string=srvban['expire_at_string'] if 'expire_at_string' in srvban else None,
                            duration_string=srvban['duration_string'] if 'duration_string' in srvban else None,
                            set_at_delta=srvban['set_at_delta'] if 'set_at_delta' in srvban else 0,
                            set_in_config=srvban['set_in_config'] if 'set_in_config' in srvban else None,
                            name=srvban['name'] if 'name' in srvban else None,
                            reason=srvban['reason'] if 'reason' in srvban else None
                        )
                )

            return self.DB_SERVERS_BANS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, type: str, name: str) -> Union[ModelServerBan, None]:
        """Retrieve all details of a single server ban (LINE).

        Args:
            type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.

        Returns:
            ModelServerBan (ModelServerBan): The model Object | None if nothing happen see Error property
        """
        try:
            response = self.Connection.query(method='server_ban.get', param={'type': type, 'name': name})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            srvban = response['result']['tkl']

            objectChannel = self.ModelServerBan(
                            type=srvban['type'] if 'type' in srvban else None,
                            type_string=srvban['type_string'] if 'type_string' in srvban else None,
                            set_by=srvban['set_by'] if 'set_by' in srvban else None,
                            set_at=srvban['set_at'] if 'set_at' in srvban else None,
                            expire_at=srvban['expire_at'] if 'expire_at' in srvban else None,
                            set_at_string=srvban['set_at_string'] if 'set_at_string' in srvban else None,
                            expire_at_string=srvban['expire_at_string'] if 'expire_at_string' in srvban else None,
                            duration_string=srvban['duration_string'] if 'duration_string' in srvban else None,
                            set_at_delta=srvban['set_at_delta'] if 'set_at_delta' in srvban else 0,
                            set_in_config=srvban['set_in_config'] if 'set_in_config' in srvban else None,
                            name=srvban['name'] if 'name' in srvban else None,
                            reason=srvban['reason'] if 'reason' in srvban else None
                        )

            return objectChannel

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add(self, type: str, name: str, reason: str, expire_at: str, duration_sting: str, _set_by: str = None) -> bool:
        """Add a server ban (LINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban#Structure_of_a_server_ban

        Args:
            type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            reason (str): The reason of the ban
            expire_at (str): Date/Time when the server ban will expire. NULL means: never.
            duration_sting (str): How long the ban will last from this point in time (human printable). Uses "permanent" for forever.
            _set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='server_ban.add', param={"type": type, "name": name, "reason": reason, "expire_at": expire_at, "duration_string": duration_sting, 'set_by': _set_by})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def del_(self, type: str, name: str, _set_by: str = None) -> bool:
        """Delete a server ban (LINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban#Structure_of_a_server_ban

        Args:
            type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            _set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='server_ban.del', param={"type": type, "name": name, 'set_by': _set_by})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class User:

    @dataclass
    class ModelUser:
        name: str
        id: str
        hostname: str
        ip: str
        details: str
        server_port: int
        client_port: int
        connected_since: str
        idle_since: str
        username: str
        realname: str
        vhost: str
        cloakedhost: str
        servername: str
        reputation: int
        security_groups: list
        modes: str
        channels: list[dict[str, str]]
        cipher: str
        certfp: str
        country_code: str
        asn: str
        asname: str

    DB_USER: list[ModelUser] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelUser], None]:
        """List users.

        Returns:
            ModelUser (list[ModelUser]): List with an object contains all Users information
        """
        try:
            response = self.Connection.query('user.list', param={'object_detail_level': 4})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            users = response['result']['list']

            for user in users:
                self.DB_USER.append(
                        self.ModelUser(
                            name=user['name'] if 'name' in user else None,
                            id=user['id'] if 'id' in user else None,
                            hostname=user['hostname'] if 'hostname' in user else None,
                            ip=user['ip'] if 'ip' in user else None,
                            details=user['details'] if 'details' in user else None,
                            server_port=user['server_port'] if 'server_port' in user else 0,
                            client_port=user['client_port'] if 'client_port' in user else 0,
                            connected_since=user['connected_since'] if 'connected_since' in user else None,
                            idle_since=user['idle_since'] if 'idle_since' in user else None,
                            username=user['user']['username'] if 'user' in user and 'username' in user['user'] else None,
                            realname=user['user']['realname'] if 'user' in user and 'realname' in user['user'] else None,
                            vhost=user['user']['vhost'] if 'user' in user and 'vhost' in user['user'] else None,
                            cloakedhost=user['user']['cloakedhost'] if 'user' in user and 'cloakedhost' in user['user'] else None,
                            servername=user['user']['servername'] if 'user' in user and 'servername' in user['user'] else None,
                            reputation=user['user']['reputation'] if 'user' in user and 'reputation' in user['user'] else 0,
                            security_groups=user['user']['security-groups'] if 'user' in user and 'security-groups' in user['user'] else [],
                            modes=user['user']['modes'] if 'user' in user and 'modes' in user['user'] else None,
                            channels=user['user']['channels'] if 'user' in user and 'channels' in user['user'] else [{}],
                            cipher=user['tls']['cipher'] if 'tls' in user and 'cipher' in user['tls'] else None,
                            certfp=user['tls']['certfp'] if 'tls' in user and 'certfp' in user['tls'] else None,
                            country_code=user['geoip']['country_code'] if 'geoip' in user and 'country_code' in user['geoip'] else None,
                            asn=user['geoip']['asn'] if 'geoip' in user and 'asn' in user['geoip'] else None,
                            asname=user['geoip']['asname'] if 'geoip' in user and 'asname' in user['geoip'] else None
                        )
                )

            return self.DB_USER

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, nickoruid: str) -> Union[ModelUser, None]:
        """Get user information

        Args:
            nickoruid (str): The nickname or uid of the user

        Returns:
            ModelUser (ModelUser): If success it return the object ModelUser

            None (None): if no value found, Probably you can see Error property
        """
        try:

            response = self.Connection.query('user.get', {'nick': nickoruid})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            user = response['result']['client']

            userObject = self.ModelUser(
                    name=user['name'] if 'name' in user else None,
                    id=user['id'] if 'id' in user else None,
                    hostname=user['hostname'] if 'hostname' in user else None,
                    ip=user['ip'] if 'ip' in user else None,
                    details=user['details'] if 'details' in user else None,
                    server_port=user['server_port'] if 'server_port' in user else 0,
                    client_port=user['client_port'] if 'client_port' in user else 0,
                    connected_since=user['connected_since'] if 'connected_since' in user else None,
                    idle_since=user['idle_since'] if 'idle_since' in user  else None,
                    username=user['user']['username'] if 'user' in user and 'username' in user['user'] else None,
                    realname=user['user']['realname'] if 'user' in user and 'realname' in user['user'] else None,
                    vhost=user['user']['vhost'] if 'user' in user and 'vhost' in user['user'] else None,
                    cloakedhost=user['user']['cloakedhost'] if 'user' in user and 'cloakedhost' in user['user'] else None,
                    servername=user['user']['servername'] if 'user' in user and 'servername' in user['user'] else None,
                    reputation=user['user']['reputation'] if 'user' in user and 'reputation' in user['user'] else 0,
                    security_groups=user['user']['security-groups'] if 'user' in user and 'security-groups' in user['user'] else [],
                    modes=user['user']['modes'] if 'user' in user and 'modes' in user['user'] else None,
                    channels=user['user']['channels'] if 'user' in user and 'channels' in user['user'] else [],
                    cipher=user['tls']['cipher'] if 'tls' in user and 'cipher' in user['tls'] else None,
                    certfp=user['tls']['certfp'] if 'tls' in user and 'certfp' in user['tls'] else None,
                    country_code=user['geoip']['country_code'] if 'geoip' in user and 'country_code' in user['geoip'] else None,
                    asn=user['geoip']['asn'] if 'geoip' in user and 'asn' in user['geoip'] else None,
                    asname=user['geoip']['asname'] if 'geoip' in user and 'asname' in user['geoip'] else None
                )

            return userObject

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def set_nick(self, nickoruid: str, newnick: str, _force: bool = False) -> bool:
        """Sets the nick name of a user (changes the nick).
        force: if set to true then q-lines (banned nick) checks will be bypassed. And also, 
        if the new nick name already exists, the other existing user is killed and we will take that new nick.

        Args:
            nick (str): the nick name or the UID
            newnick (str): the new nick name
            force (bool, optional): q-lines (banned nick) checks will be bypassed. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_nick', {'nick': nickoruid, 'newnick': newnick, 'force': _force})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_username(self, nickoruid: str, username:str) -> bool:
        """Set the username / ident of a user.

        Args:
            nickoruid (str): the nick name or the UID
            username (str): the new user name / ident

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_username', {'nick': nickoruid, 'username': username})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_realname(self, nickoruid: str, realname:str) -> bool:
        """Set the realname / gecos of a user.

        Args:
            nickoruid (str): the nick name or the UID
            realname (str):  the new realname / gecos of a user.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_realname', {'nick': nickoruid, 'realname': realname})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_vhost(self, nickoruid: str, vhost:str) -> bool:
        """Set a virtual host (vhost) on the user.

        Args:
            nickoruid (str): the nick name or the UID
            vhost (str): the new virtual host.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_vhost', {'nick': nickoruid, 'vhost': vhost})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_mode(self, nickoruid: str, modes:str) -> bool:
        """Change the modes of a user.

        Args:
            nickoruid (str): the nick name or the UID
            modes (str): the mode string, eg -i+w

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_mode', {'nick': nickoruid, 'modes': modes})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_snomask(self, nickoruid: str, snomask:str) -> bool:
        """Change the snomask of a user.

        Args:
            nickoruid (str): the nick name or the UID
            snomask (str): the snomask string, eg -j+R

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_snomask', {'nick': nickoruid, 'snomask': snomask})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def set_oper(self, nickoruid: str, oper_account:str, oper_class: str, _class: str = '', _modes: str = '', _snomask: str = '', _vhost: str = '') -> bool:
        """Make user an IRC operator.

        Args:
            nickoruid (str): the nick name or the UID
            oper_account (str):  the oper account, to be shown in WHOIS to fellow ircops and in logs.
            oper_class (str):  the operclass. Usually one of the default operclasses like netadmin-with-override
            _class (str, optional): the class to put the user in. If this option is not specified then opers is assumed, since this class exists in most unrealircd.conf's. You can specify "" (empty) if you don't want to put the user in a class. Defaults to ''.
            _modes (str, optional): user modes to set on oper. For example: +xws. If this option is not specified then set::modes-on-oper is used. You can specify "" (empty) if you don't want to set any additional modes on the user. Defaults to ''.
            _snomask (str, optional): snomask to set on oper. For example: +bBcdfkqsSoO. If this option is not specified then set::snomask-on-oper is used. You can specify "" (empty) if you don't want to set any snomasks on the user. Defaults to ''.
            _vhost (str, optional): virtual host to set on oper. Defaults to ''.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.set_oper', {'nick': nickoruid, 'oper_account': oper_account, 'oper_class': oper_class, 'class': _class, 'modes': _modes, 'snomask': _snomask, 'vhost': _vhost})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def join(self, nickoruid: str, channel:str, _key: str = '', _force: bool = False) -> bool:
        """Join a user to a channel.

        Note: If force is set to true then the user will walk through bans, modes and other restrictions 
        (similar to SAJOIN on IRC). If force is not set, or set to false, then it will be a regular JOIN attempt that may fail. 
        If it fails then the user will see an error message on their screen, such as for example Cannot join channel (+k).

        Args:
            nickoruid (str): the nick name or the UID
            channel (str): the channel(s) to join (e.g. #channel or #channel1,#channel2)
            _key (str, optional): the key of the channel(s) (only for channels with +k needed, again separate by colon for multiple channels). Defaults to ''.
            _force (bool, optional): whether to bypass join restrictions or not. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.join', {'nick': nickoruid, 'channel': channel, 'key': _key, 'force': _force})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def part(self, nickoruid: str, channel:str, _force: bool = False) -> bool:
        """Part a user from a channel.

        Note: If force is set to true then the user will see a notice that they were forcefully PARTed from the channel(s). 
        If force is not set, or set to false, then there will be no such notice.

        Args:
            nickoruid (str): the nick name or the UID
            channel (str): the channel(s) to join (e.g. #channel or #channel1,#channel2)
            _force (bool, optional): whether to bypass join restrictions or not. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.part', {'nick': nickoruid, 'channel': channel, 'force': _force})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def kill(self, nickoruid: str, reason:str) -> bool:
        """Kill a user, showing that the user was forcefully removed.

        Note: There is also user.quit. The difference is that user.kill will show that the user is forcefully killed, while user.quit pretend the user quit normally. 
        This slight difference may invoke certain client behavior. For example, some clients don't immediately reconnect for a KILL but do so on a QUIT.

        Args:
            nickoruid (str): the nick name or the UID
            reason (str): reason for the kill

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.kill', {'nick': nickoruid, 'reason': reason})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def quit(self, nickoruid: str, reason:str) -> bool:
        """Quit a user, pretending it was a normal QUIT.

        Note: There is also user.kill. The difference is that user.kill will show that the user is forcefully killed, 
        while user.quit pretend the user quit normally. This slight difference may invoke certain client behavior. 
        For example, some clients don't immediately reconnect for a KILL but do so on a QUIT.

        Args:
            nickoruid (str): the nick name or the UID
            reason (str): reason for the kill

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            response = self.Connection.query('user.quit', {'nick': nickoruid, 'reason': reason})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Stats:

    @dataclass
    class ModelStats:
        server_total: int
        server_ulined: int
        user_total: int
        user_ulined: int
        user_oper: int
        user_record: int
        user_countries: list[dict[str, any]]
        channel_total: int
        server_ban_total: int
        server_ban_server_ban: int
        server_ban_spamfilter: int
        server_ban_name_ban: int
        server_ban_server_ban_exception: int

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def get(self, _object_detail_level:int = 1) -> Union[ModelStats, None]:
        """Get some quick basic statistics.

        Args:
            _object_detail_level (int, optional): set the detail of the response object. The user.countries will be included from level 1+. The default object_detail_level is 1. Defaults to 1.

        Returns:
            list[ModelStats]: The Stats model or None if there is an error see Error Property
        """
        try:
            response = self.Connection.query('stats.get', param={'object_detail_level': _object_detail_level})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            stats = response['result']

            statsModel = self.ModelStats(
                    server_total=stats['server']['total'] if 'server' in stats and 'total' in stats['server'] else 0,
                    server_ulined=stats['server']['ulined'] if 'server' in stats and 'ulined' in stats['server'] else 0,

                    user_total=stats['user']['total'] if 'user' in stats and 'total' in stats['user'] else 0,
                    user_ulined=stats['user']['ulined'] if 'user' in stats and 'total' in stats['user'] else 0,
                    user_oper=stats['user']['oper'] if 'user' in stats and 'oper' in stats['user'] else 0,
                    user_record=stats['user']['record'] if 'user' in stats and 'record' in stats['user'] else 0,
                    user_countries=stats['user']['countries'] if 'user' in stats and 'countries' in stats['user'] else [],
                    
                    channel_total=stats['channel']['total'] if 'channel' in stats and 'total' in stats['channel'] else 0,

                    server_ban_total=stats['server_ban']['total'] if 'server_ban' in stats and 'total' in stats['server_ban'] else 0,
                    server_ban_server_ban=stats['server_ban']['server_ban'] if 'server_ban' in stats and 'server_ban' in stats['server_ban'] else 0,
                    server_ban_spamfilter=stats['server_ban']['spamfilter'] if 'server_ban' in stats and 'spamfilter' in stats['server_ban'] else 0,
                    server_ban_name_ban=stats['server_ban']['name_ban'] if 'server_ban' in stats and 'name_ban' in stats['server_ban'] else 0,
                    server_ban_server_ban_exception=stats['server_ban']['server_ban_exception'] if 'server_ban' in stats and 'server_ban_exception' in stats['server_ban'] else 0
                )

            return statsModel

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')





import json.scanner
import requests, json, urllib3, socket, re, sys, os
from requests.auth import HTTPBasicAuth
import base64, ssl, time, logging, random
from typing import Literal, Union
from types import SimpleNamespace
from dataclasses import dataclass

class Connection:

    @dataclass
    class ErrorModel:
        code: int
        message: str

    def __init__(self, req_method:str, url: str, path_to_socket_file: str, username: str, password: str, debug_level: int) -> None:

        self.debug_level = debug_level
        self.Logs: logging
        self.__init_log_system()
        self.Error = self.ErrorModel(0, '')

        self.url = url
        self.path_to_socket_file = path_to_socket_file

        self.endpoint: str = None
        self.host: str = None
        self.port: int = 0
        self.username = username
        self.password = password

        self.request: str = ''
        self.req_method = req_method
        self.str_response = ''
        self.json_response = ''

        # Option 2 with Namespacescs
        self.json_response_np: SimpleNamespace
        self.query('stats.get')

    def __check_url(self, url: str) -> bool:
        """Check provided url if it follow the format

        Args:
            url (str): Url to jsonrpc https://your.rpcjson.link:port/api

        Returns:
            bool: True if url is correct else False
        """
        try:
            response = False

            if url is None:
                return response

            pattern = r'https?://([a-zA-Z0-9\.-]+):(\d+)/(.+)'

            match = re.match(pattern, url)

            if not match is None:
                self.host = match.group(1)
                self.port = match.group(2)
                self.endpoint = match.group(3)
                response = True
            else:
                self.Error.code = -1
                self.Error.message = 'You must provide the url in this format: https://your.rpcjson.link:port/api'

            return response
        except NameError as nameerr:
            self.Logs.critical(f'NameError: {nameerr}')

    def __check_unix_socket_file(self, path_to_socket_file: str) -> bool:
        """Check provided full path to socket file if it exist

        Args:
            path_to_socket_file (str): Full path to unix socket file

        Returns:
            bool: True if path is correct else False
        """
        try:
            response = False

            if path_to_socket_file is None:
                self.Error.code = -1
                self.Error.message = 'The Path to your socket file is empty ? please be sure that you are providing the correct socket path'
                return response

            if not os.path.exists(path_to_socket_file):
                self.Error.code = -1
                self.Error.message = 'The Path to your socket file is wrong ? please make sure that you are providing the correct socket path'
                return response

            response = True

            return response
        except NameError as nameerr:
            self.Logs.critical(f'NameError: {nameerr}')

    def __is_error_connection(self, response: str) -> bool:
        """If True, it means that there is an error

        Args:
            response (str): The response to analyse

        Returns:
            bool: True if there is a connection error
        """
        if 'authentication required' == response.lower().strip():
            self.Error.code = -1
            self.Error.message = '>> Authentication required'
            return True
        else:
            return False

    def __send_to_unixsocket(self):
        try:

            sock = socket.socket(socket.AddressFamily.AF_UNIX, socket.SocketKind.SOCK_STREAM)

            if not self.__check_unix_socket_file(self.path_to_socket_file):
                return None

            sock.connect(self.path_to_socket_file)
            sock.settimeout(10)

            if not self.request:
                return None

            sock.sendall(f'{self.request}\r\n'.encode())

            response = b""
            chunk = b""
            pattern = b'\n$'

            while True:
                chunk = sock.recv(4096)
                response += chunk
                if re.findall(pattern, chunk):
                    break

            str_data = response.decode()
            self.json_response = json.loads(str_data)
            self.json_response_np: SimpleNamespace = json.loads(str_data, object_hook=lambda d: SimpleNamespace(**d))

            sock.close()

        except AttributeError as attrerr:
            self.Logs.critical(f'AF_Unix Error: {attrerr}')
            self.Error.code = -1
            self.Error.message = 'AF_UNIX Are you sure you want to use Unix socket ?'
        except OSError as oserr:
            self.Logs.critical(f'System Error: {oserr}')
        except Exception as err:
            self.Logs.error(f'General Error: {err}')

    def __send_srequest(self):
        """S For socket connection"""
        try:

            if not self.__check_url(self.url) and not self.url is None:
                self.Logs.critical('You must provide the url in this format: https://your.rpcjson.link:port/api')
                return None

            get_url = self.url
            get_host = self.host
            get_port = self.port
            get_username = self.username
            get_password = self.password
            credentials = base64.b64encode(f"{get_username}:{get_password}".encode()).decode()

            # Create a socket and wrap it in an SSL context for HTTPS
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

            with socket.create_connection((get_host, get_port)) as sock:
                with context.wrap_socket(sock, server_hostname=get_host) as ssock:
                    # Send the HTTPS

                    json_request = self.request
                    request = self.__build_headers(credentials, str(json_request))
                    ssock.sendall(request.encode('utf-8'))

                    # Receive the response
                    response = b""
                    while True:
                        chunk = ssock.recv(4096)
                        if not chunk:
                            break
                        response += chunk

                    # Decode and print the response
                    response_str = response.decode('utf-8')
                    header_end = response_str.find("\r\n\r\n")
                    if header_end != -1:
                        body = response_str[header_end + 4:]  # Extract the body
                    else:
                        body = response_str

                    if self.__is_error_connection(body):
                        return None

                    self.json_response = json.loads(body)
                    self.json_response_np: SimpleNamespace = json.loads(body, object_hook=lambda d: SimpleNamespace(**d))

        except (socket.error, ssl.SSLError) as serror:
            self.Logs.error(f'Socket Error: {serror}')
        except Exception as err:
            self.Logs.error(f'General Error: {err}')
            # self.Logs.error(f'General Error: {traceback.format_exc()}')

    def __send_request(self) :
        """Use requests module"""
        try:

            if not self.__check_url(self.url) and not self.url is None:
                self.Logs.critical('You must provide the url in this format: https://your.rpcjson.link:port/api')
                return None

            verify = False

            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

            credentials = HTTPBasicAuth(self.username, self.password)
            jsonrequest = self.request

            response = requests.post(url=self.url, auth=credentials, data=jsonrequest, verify=verify)

            if self.__is_error_connection(response.text):
                return None

            decodedResponse = json.dumps(response.text)

            self.str_response = decodedResponse
            self.json_response = json.loads(json.loads(decodedResponse))
            self.json_response_np: SimpleNamespace = json.loads(response.text, object_hook=lambda d: SimpleNamespace(**d))

        except KeyError as ke:
            self.Logs.error(f"KeyError : {ke}")
            self.Logs.error(f"Initial request: {self.request}")
        except requests.ReadTimeout as rt:
            self.Logs.error(f"Timeout : {rt}")
            self.Logs.error(f"Initial request: {self.request}")
        except requests.ConnectionError as ce:
            self.Logs.critical(f"Connection Error : {ce}")
            self.Logs.critical(f"Initial request: {self.request}")
        except json.decoder.JSONDecodeError as jsonerror:
            self.Logs.error(f"jsonError {jsonerror}")
            self.Logs.error(f"Initial request: {self.request}")
        except Exception as err:
            self.Logs.error(f"General Error {err}")
            self.Logs.error(f"Initial request: {self.request}")

    def __build_headers(self, credentials: str, data:str) -> str:
        """Build the header for socket connection only

        Args:
            credentials (str): crypted credentials
            data (str): data we need to send to the server

        Returns:
            str: Returd the built header
        """
        headers: str = (
                    f"POST /{self.endpoint} HTTP/1.1\r\n"
                    f"Host: {self.host}\r\n"
                    "Content-Type: application/json\r\n"
                    f"Authorization: Basic {credentials}\r\n"
                    f"Content-Length: {len(data)}\r\n"
                    "Connection: close\r\n"
                    "\r\n"
                    f"{data}"
                )

        return headers

    def __init_log_system(self) -> None:
        """Init log system
        """
        # Init logs object
        self.Logs = logging
        self.Logs.basicConfig(level=self.debug_level,
                            encoding='UTF-8',
                            format='%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(funcName)s - %(message)s')

    def query(self, method: Union[Literal['stats.get', 'rpc.info','user.list'], str], param: dict = {}, id: int = 123, jsonrpc:str = '2.0') -> Union[str, any, None, bool]:
        """This method will use to run the queries

        Args:
            method (Union[Literal[&#39;stats.get&#39;, &#39;rpc.info&#39;,&#39;user.list&#39;], str]): The method to send to unrealircd
            param (dict, optional): the paramaters to send to unrealircd. Defaults to {}.
            id (int, optional): id of the request. Defaults to 123.
            jsonrpc (str, optional): jsonrpc. Defaults to '2.0'.

        Returns:
            str: The correct response
            None: no response from the server
            bool: False if there is an error occured
        """

        # data = '{"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 123}'
        get_method = method
        get_param = param

        if id == 123:
            rand = random.randint(1, 6000)
            get_id = int(time.time()) + rand

        else:
            get_id = id

        response = {
            "jsonrpc": jsonrpc,
            "method": get_method,
            "params": get_param,
            "id": get_id
        }

        self.request = json.dumps(response)

        if self.req_method == 'socket':
            self.__send_srequest()
        elif self.req_method == 'unixsocket':
            self.__send_to_unixsocket()
        elif self.req_method == 'requests':
            self.__send_request()
        else:
            print('No valid request method')

        if self.json_response == '':
            return False

        return self.json_response

    def set_error(self, json_error: dict):
        """set the error

        Args:
            code (int): the error code
            message (str): the message of the error
        """
        try:
            if 'error' in json_error:
                self.Error.code=json_error['error']['code']
                self.Error.message=json_error['error']['message']

        except KeyError as ke:
            self.Logs.error(ke)
        except Exception as err:
            self.Logs.error(err)


from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Rpc:

    @dataclass
    class ModelRpcInfo:
        name: str
        module: str
        version: str

    DB_RPC_INFO: list[ModelRpcInfo] = []


    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def info(self) -> Union[list[ModelRpcInfo], None]:
        """A response object, with in the result object a "methods" object which is a list of: the API method with in that the name, module name and module version.

        Returns:
            list[ModelRpcInfo]: List of ModelRpcInfo, None if nothing see the Error property
        """
        try:
            response = self.Connection.query(method='rpc.info')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            rpcinfos = response['result']['methods']

            for rpcinfo in rpcinfos:
                self.DB_RPC_INFO.append(
                        self.ModelRpcInfo(
                            name=rpcinfos[rpcinfo]['name'],
                            module=rpcinfos[rpcinfo]['module'],
                            version=rpcinfos[rpcinfo]['version'],
                            )
                )

            return self.DB_RPC_INFO

        except KeyError as ke:
            self.Logs.error(ke)
        except Exception as err:
            self.Logs.error(err)

    def set_issuer(self, name: str) -> bool:
        """Set who is the issuer of all the subsequent commands that are done. 

        This information will be used by the logging system and communicate to servers and IRCOps in the unrealircd.org/issued-by message tag.

        This method only exists in UnrealIRCd 6.1.0 and later

        Args:
            name (str): the name of the person who is issuing the commands, eg. the current logged in person to the admin panel. Note: This name must conform to the regular nickname rules, eg. may not contain spaces and the like.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='rpc.set_issuer', 
                param={"name": name}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add_timer(self, timer_id: str, every_msec: int, request: dict) -> bool:
        """Add a timer so a JSON-RPC request is executed at certain intervals.

        This method only exists in UnrealIRCd 6.1.0 and later

        exemple of request:

        {"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 555}

        Args:
            timer_id (str): the name/id of the timer, this so you can cancel it via rpc.del_timer later. It must be unique for this RPC connection.
            every_msec (int): the timer will be executed every -this- milliseconds. The minimum value is 250 (a quarter of a second).
            request (dict): the full JSON-RPC request, so {"jsonrpc":"2.0" etc. etc.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='rpc.add_timer',
                param={"timer_id": timer_id, "every_msec": every_msec, "request": request}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def del_timer(self, timer_id: str) -> bool:
        """Remove a previously added timer. Note that you can only cancel timers that belong to your own connection.

        This method only exists in UnrealIRCd 6.1.0 and later

        Args:
            timer_id (str): the name/id of the timer that needs to be canceled.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='rpc.del_timer', 
                param={"timer_id": timer_id}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Log:

    @dataclass
    class ModelLog:
        type: str
        type_string: str
        set_by: str
        set_at: str
        expire_at: str
        set_at_string: str
        expire_at_string: str
        duration_string: str
        set_at_delta: int
        set_in_config: bool
        name: str
        reason: str

    DB_NAME_BANS: list[ModelLog] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self, sources: list = None) -> Union[SimpleNamespace, None]:
        """Fetch past log entries (since boot).

        Returns:
            list[ModelNameBan]: List of ModelNameBan, None if nothing or Error see the property
        """
        try:
            response = self.Connection.query(
                method='log.list', 
                param={"sources": sources}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            return self.response_np

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def subscribe(self, sources: list) -> bool:
        """Subscribe to one or more log sources. Any previous subscriptions are overwritten (lost).

        Args:
            sources (list): an array of log sources. See log block sources. 
            For example ["!debug","all"] would give you all log messages except for debug messages. And ["connect"] would give only client connects/disconnects.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='log.subscribe', 
                param={"sources": sources}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True
                else:
                    self.Logs.debug(response)
                    return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def unsubscribe(self) -> bool:
        """Unsubscribe from all log events.
        After this UnrealIRCd will stop streaming log events to you.

        Returns:
            bool: Always returns true
        """
        try:
            response = self.Connection.query(
                method='log.unsubscribe'
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True
                else:
                    self.Logs.debug(response)
                    return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Server_ban_exception:

    @dataclass
    class ModelServerBanException:
        type: str
        type_string: str
        set_by: str
        set_at: str
        expire_at: str
        set_at_string: str
        expire_at_string: str
        duration_string: str
        set_at_delta: int
        set_in_config: bool
        name: str
        reason: str
        exception_types: str

    DB_SERVERS_BANS_EXCEPTION: list[ModelServerBanException] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelServerBanException], None]:
        """List server ban exceptions (ELINEs).

        Returns:
            list[ModelServerBanException]: List of ModelServerBanException, None if nothing see the Error property
        """
        try:
            response = self.Connection.query(method='server_ban_exception.list')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            srvbansexceps = response['result']['list']

            for srvbansexcep in srvbansexceps:
                self.DB_SERVERS_BANS_EXCEPTION.append(
                        self.ModelServerBanException(
                            type=srvbansexcep['type'] if 'type' in srvbansexcep else None,
                            type_string=srvbansexcep['type_string'] if 'type_string' in srvbansexcep else None,
                            set_by=srvbansexcep['set_by'] if 'set_by' in srvbansexcep else None,
                            set_at=srvbansexcep['set_at'] if 'set_at' in srvbansexcep else None,
                            expire_at=srvbansexcep['expire_at'] if 'expire_at' in srvbansexcep else None,
                            set_at_string=srvbansexcep['set_at_string'] if 'set_at_string' in srvbansexcep else None,
                            expire_at_string=srvbansexcep['expire_at_string'] if 'expire_at_string' in srvbansexcep else None,
                            duration_string=srvbansexcep['duration_string'] if 'duration_string' in srvbansexcep else None,
                            set_at_delta=srvbansexcep['set_at_delta'] if 'set_at_delta' in srvbansexcep else 0,
                            set_in_config=srvbansexcep['set_in_config'] if 'set_in_config' in srvbansexcep else None,
                            name=srvbansexcep['name'] if 'name' in srvbansexcep else None,
                            reason=srvbansexcep['reason'] if 'reason' in srvbansexcep else None,
                            exception_types=srvbansexcep['exception_types'] if 'exception_types' in srvbansexcep else None
                        )
                )

            return self.DB_SERVERS_BANS_EXCEPTION

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, name: str) -> Union[ModelServerBanException, None]:
        """Retrieve all details of a single server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.

        Returns:
            ModelServerBanException (ModelServerBanException): The model ModelServerBanException Object | None if nothing see the Error property
        """
        try:
            response = self.Connection.query(method='server_ban_exception.get', param={'name': name})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            srvbanexcep = response['result']['tkl']

            objectServBanExcep = self.ModelServerBanException(
                            type=srvbanexcep['type'] if 'type' in srvbanexcep else None,
                            type_string=srvbanexcep['type_string'] if 'type_string' in srvbanexcep else None,
                            set_by=srvbanexcep['set_by'] if 'set_by' in srvbanexcep else None,
                            set_at=srvbanexcep['set_at'] if 'set_at' in srvbanexcep else None,
                            expire_at=srvbanexcep['expire_at'] if 'expire_at' in srvbanexcep else None,
                            set_at_string=srvbanexcep['set_at_string'] if 'set_at_string' in srvbanexcep else None,
                            expire_at_string=srvbanexcep['expire_at_string'] if 'expire_at_string' in srvbanexcep else None,
                            duration_string=srvbanexcep['duration_string'] if 'duration_string' in srvbanexcep else None,
                            set_at_delta=srvbanexcep['set_at_delta'] if 'set_at_delta' in srvbanexcep else 0,
                            set_in_config=srvbanexcep['set_in_config'] if 'set_in_config' in srvbanexcep else None,
                            name=srvbanexcep['name'] if 'name' in srvbanexcep else None,
                            reason=srvbanexcep['reason'] if 'reason' in srvbanexcep else None,
                            exception_types=srvbanexcep['exception_types'] if 'exception_types' in srvbanexcep else None
                        )

            return objectServBanExcep

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add(self, name: str, exception_types: str, reason: str, _set_by: str = None, _expire_at: str = None, _duration_sting: str = None) -> bool:
        """Add a server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): user@host mask or extended server ban
            exception_types (str): eg k for a kline exception
            reason (str): The reason of the ban
            expire_at (str, optional): Date/Time when the server ban will expire. NULL means: never. Defaults to None.
            duration_sting (str, optional): How long the ban will last from this point in time (human printable). Uses "permanent" for forever. Defaults to None.
            _set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='server_ban_exception.add', param={"name": name, "exception_types": exception_types, "reason": reason, "expire_at": _expire_at, "duration_string": _duration_sting, 'set_by': _set_by})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def del_(self, name: str, _set_by: str = None) -> bool:
        """Delete a server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            _set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(method='server_ban_exception.del', param={"name": name, 'set_by': _set_by})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from typing import Literal
from unrealircd_rpc_py.Connection import Connection
from unrealircd_rpc_py.User import User
from unrealircd_rpc_py.Stats import Stats
from unrealircd_rpc_py.Whowas import Whowas
from unrealircd_rpc_py.Server import Server
from unrealircd_rpc_py.Channel import Channel
from unrealircd_rpc_py.Server_ban import Server_ban
from unrealircd_rpc_py.Server_ban_exeption import Server_ban_exception
from unrealircd_rpc_py.Spamfilter import Spamfilter
from unrealircd_rpc_py.Name_ban import Name_ban
from unrealircd_rpc_py.Rpc import Rpc
from unrealircd_rpc_py.Log import Log

class Loader:

    def __init__(self, req_method: Literal['requests', 'socket', 'unixsocket'], url: str = None, path_to_socket_file: str = None, username: str = None, password: str = None, debug_level: Literal[10, 20, 30, 40, 50] = 20) -> None:
        """Initiate connection to unrealircd

        requests and socket:
        If you use request or socket as req_method you must provide:
            \n url, username and password

        unixsocket:
        If you use request or socket as req_method you must provide:
            \n path_to_socket_file: /path/to/unrealircd/socket/your_socket.socket

        ## Exemples:
        If you want to use unix socket you need to provide 2 parameters:
            \n Loader(
                req_method='unixsocket',
                path_to_socket_file='/path/to/unrealircd/socket/your_socket.socket'
                )

        If you want to use remote connexion you need to provide 4 parameters:
            \n Loader(
                req_method='requests',
                url='https://your.rpcjson.link:port/api',
                username='readonly',
                password='azerty'
                )

        Args:
            req_method (str): The method you want to use, 3 options are available requests | socket | unixsocket
            url (str, optional): The full url to connect https://your.rpcjson.link:port/api.
            path_to_socket_file (str, optional): The full path to your unix socket file
            username (str, optional): Default to None 
            password (str, optional): Default to None 
            debug_level (str, optional): NOTSET=0 | DEBUG=10 | INFO=20 | WARN=30 | ERROR=40 | CRITICAL=50 - Default to 20 
        """
        self.Connection = Connection(
            url=url,
            path_to_socket_file=path_to_socket_file,
            username=username,
            password=password,
            req_method=req_method,
            debug_level=debug_level
        )

        # Create ErrorModel Instance
        self.Error = self.Connection.Error
        """Access Error Object when there is an error"""

        # Create User Instance
        self.User = User(self.Connection)
        """The User module instance"""

        # Create Server Instance
        self.Server = Server(self.Connection)
        """The Server module instance"""

        # Create Server_ban Instance
        self.Server_ban = Server_ban(self.Connection)
        """The Server_ban module instance"""

        # Create Server_ban_exception Instance
        self.Server_ban_exception = Server_ban_exception(self.Connection)
        """The Server_ban_exception module instance"""

        # Create Name_ban Instance
        self.Name_ban = Name_ban(self.Connection)
        """The Name_ban module instance"""

        # Create Rpc Instance
        self.Rpc = Rpc(self.Connection)
        """The Rpc module instance"""

        # Create Spamfilter Instance
        self.Spamfilter = Spamfilter(self.Connection)
        """The Spamfilter module instance"""

        # Create Channel Instance
        self.Channel = Channel(self.Connection)
        """The Channel module instance"""

        # Create Stats Instance
        self.Stats = Stats(self.Connection)
        """The Stats module instance"""

        # Create Whowas Instance
        self.Whowas = Whowas(self.Connection)
        """The Whowas module instance"""



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Whowas:

    @dataclass
    class ModelWhowas:
        name: str
        event: str
        logon_time: str
        logoff_time: str
        hostname: str
        ip: str
        details: str
        connected_since: str
        user_username: str
        user_realname: str
        user_vhost: str
        user_servername: str
        geoip_country_code: str
        geoip_asn: str
        geoip_asname: str

    DB_WHOWAS: list[ModelWhowas] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def get(self, _nick: str = None, _ip: str = None, _object_detail_level:int = 2) -> Union[list[ModelWhowas], None]:
        """Get WHOWAS history of a user. 6.1.0+

        Args:
            _nick (str, optional): The nick name to search for. Defaults to None.
            _ip (str, optional): The IP address to search for. Defaults to None.
            _object_detail_level (int, optional): set the detail of the response object, this is similar to the Detail level column in Structure of a client object. In this RPC call it defaults to 2 if this parameter is not specified. Defaults to 2.

        Returns:
            List[ModelWhowas]: If success it return the object ModelWhowas | None if nothing or error, see Error property
        """
        try:
            response = self.Connection.query('whowas.get', param={'nick': _nick, 'ip': _ip, 'object_detail_level': _object_detail_level})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            whowass = response['result']['list']

            for whowas in whowass:
                self.DB_WHOWAS.append(
                    self.ModelWhowas(
                        name=whowas['name'] if 'name' in whowas else None,
                        event=whowas['event'] if 'event' in whowas else None,
                        logon_time=whowas['logon_time'] if 'logon_time' in whowas else None,
                        logoff_time=whowas['logoff_time'] if 'logoff_time' in whowas else None,
                        hostname=whowas['hostname'] if 'hostname' in whowas else None,
                        ip=whowas['ip'] if 'ip' in whowas else None,
                        details=whowas['details'] if 'details' in whowas else None,
                        connected_since=whowas['connected_since'] if 'connected_since' in whowas else None,
                        user_username=whowas['user'] if 'user' in whowas and 'username' in whowas['user'] else None,
                        user_realname=whowas['user'] if 'user' in whowas and 'realname' in whowas['user'] else None,
                        user_vhost=whowas['user'] if 'user' in whowas and 'vhost' in whowas['user'] else None,
                        user_servername=whowas['user'] if 'user' in whowas and 'servername' in whowas['user'] else None,
                        geoip_country_code=whowas['geoip'] if 'geoip' in whowas and 'username' in whowas['geoip'] else None,
                        geoip_asn=whowas['geoip'] if 'geoip' in whowas and 'asn' in whowas['geoip'] else None,
                        geoip_asname=whowas['geoip'] if 'geoip' in whowas and 'asname' in whowas['geoip'] else None
                        )
                    )

            return self.DB_WHOWAS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



from setuptools import setup, find_packages

setup(
    name='unrealircd_rpc_py',
    version='0.1.2',
    packages=find_packages(),
    install_requires=[
        "requests>=2.25.1"
    ],
    author='adator',
    author_email='debian@deb.biz.st',
    description='Python library for unrealIRCD json-rpc',
    long_description=open('README.MD').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/adator85/unrealircd_rpc_py',
    package_data={
        '': ['README.MD'],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.10'
)


from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Server:

    @dataclass
    class ModelServer:
        name: str
        id: str
        hostname: str
        ip: str
        details: str
        server_port: int
        client_port: int
        connected_since: str
        idle_since: str

        server_info: str
        server_uplink: str
        server_num_users: int
        server_boot_time: str
        server_synced: bool
        server_ulined: bool

        server_features_software: str
        server_features_protocol: int
        server_features_usermodes: str
        server_features_chanmodes: list[str]
        server_features_rpc_modules: list[dict[str, any]]

        tls_cipher: str
        tls_certfp: str
   
    @dataclass
    class ModelRehash:
        name: str
        id: str
        hostname: str
        ip: str
        server_port: int
        details: str
        connected_since: str
        idle_since: str
        log: list[dict]
        success: bool

    @dataclass
    class ModelModules:
        name: str
        version: str
        author: str
        description: str
        third_party: bool
        permanent: bool
        permanent_but_reloadable: bool

    DB_SERVER: list[ModelServer] = []
    DB_MODULES: list[ModelModules] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelServer], None]:
        """List servers.

        Returns:
            list[ModelServer]: List with an object contains all Servers information
        """
        try:
            response = self.Connection.query('server.list')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            servers = response['result']['list']

            for server in servers:
                self.DB_SERVER.append(
                        self.ModelServer(
                            name=server['name'] if 'name' in server else None,
                            id=server['id'] if 'id' in server else None,
                            hostname=server['hostname'] if 'hostname' in server else None,
                            ip=server['ip'] if 'ip' in server else None,
                            details=server['details'] if 'details' in server else None,
                            server_port=server['server_port'] if 'server_port' in server else 0,
                            client_port=server['client_port'] if 'client_port' in server else 0,
                            connected_since=server['connected_since'] if 'connected_since' in server else None,
                            idle_since=server['idle_since'] if 'idle_since' in server else None,

                            server_info=server['server']['info'] if 'server' in server and 'info' in server['server'] else None,
                            server_uplink=server['server']['uplink'] if 'server' in server and 'uplink' in server['server'] else None,
                            server_num_users=server['server']['num_users'] if 'server' in server and 'num_users' in server['server'] else 0,
                            server_boot_time=server['server']['boot_time'] if 'server' in server and 'boot_time' in server['server'] else None,
                            server_synced=server['server']['synced'] if 'server' in server and 'synced' in server['server'] else False,
                            server_ulined=server['server']['ulined'] if 'server' in server and 'ulined' in server['server'] else False,

                            server_features_software=server['server']['features']['software'] if 'server' in server and 'features' in server['server'] and 'software' in server['server']['features'] else None,
                            server_features_protocol=server['server']['features']['protocol'] if 'server' in server and 'features' in server['server'] and 'protocol' in server['server']['features'] else 0,
                            server_features_usermodes=server['server']['features']['usermodes'] if 'server' in server and 'features' in server['server'] and 'usermodes' in server['server']['features'] else None,
                            server_features_chanmodes=server['server']['features']['chanmodes'] if 'server' in server and 'features' in server['server'] and 'chanmodes' in server['server']['features'] else [],
                            server_features_rpc_modules=server['server']['features']['rpc_modules'] if 'server' in server and 'features' in server['server'] and 'rpc_modules' in server['server']['features'] else [],

                            tls_cipher=server['tls']['cipher'] if 'tls' in server and 'cipher' in server['tls'] else None,
                            tls_certfp=server['tls']['certfp'] if 'tls' in server and 'certfp' in server['tls'] else None
                        )
                )

            return self.DB_SERVER

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, _serverorsid: str = None) -> Union[ModelServer, None]:
        """Retrieve all details of a single server.

        Args:
            _server (str, optional): the server name (or the SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API). Defaults to None.

        Returns:
            Union[ModelServer, None, bool]: The ModelServer if success | None if nothing | False if error
        """
        try:

            response = self.Connection.query('server.get', {'server': _serverorsid})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            server = response['result']['server']

            serverObject = self.ModelServer(
                    name=server['name'] if 'name' in server else None,
                    id=server['id'] if 'id' in server else None,
                    hostname=server['hostname'] if 'hostname' in server else None,
                    ip=server['ip'] if 'ip' in server else None,
                    details=server['details'] if 'details' in server else None,
                    server_port=server['server_port'] if 'server_port' in server else 0,
                    client_port=server['client_port'] if 'client_port' in server else 0,
                    connected_since=server['connected_since'] if 'connected_since' in server else None,
                    idle_since=server['idle_since'] if 'idle_since' in server else None,

                    server_info=server['server']['info'] if 'server' in server and 'info' in server['server'] else None,
                    server_uplink=server['server']['uplink'] if 'server' in server and 'uplink' in server['server'] else None,
                    server_num_users=server['server']['num_users'] if 'server' in server and 'num_users' in server['server'] else None,
                    server_boot_time=server['server']['boot_time'] if 'server' in server and 'boot_time' in server['server'] else None,
                    server_synced=server['server']['synced'] if 'server' in server and 'synced' in server['server'] else None,
                    server_ulined=server['server']['ulined'] if 'server' in server and 'ulined' in server['server'] else None,

                    server_features_software=server['server']['features']['software'] if 'server' in server and 'features' in server['server'] and 'software' in server['server']['features'] else None,
                    server_features_protocol=server['server']['features']['protocol'] if 'server' in server and 'features' in server['server'] and 'protocol' in server['server']['features'] else None,
                    server_features_usermodes=server['server']['features']['usermodes'] if 'server' in server and 'features' in server['server'] and 'usermodes' in server['server']['features'] else None,
                    server_features_chanmodes=server['server']['features']['chanmodes'] if 'server' in server and 'features' in server['server'] and 'chanmodes' in server['server']['features'] else None,
                    server_features_rpc_modules=server['server']['features']['rpc_modules'] if 'server' in server and 'features' in server['server'] and 'rpc_modules' in server['server']['features'] else None,

                    tls_cipher=server['tls']['cipher'] if 'tls' in server and 'cipher' in server['tls'] else None,
                    tls_certfp=server['tls']['certfp'] if 'tls' in server and 'certfp' in server['tls'] else None
                )

            return serverObject

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def rehash(self, _serverorsid: str = None) -> Union[ModelRehash, None]:
        """Rehash the server.

        IMPORTANT: If all servers on your network have rpc.modules.default.conf included then this can return an object with full rehash details.
        Otherwise, for remote rehashes a simple boolean 'true' result is returned.

        Args:
            _serverorsid (str, optional): the server name (or SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API). Defaults to None.

        Returns:
            ModelRehash: True if success or False if failed
        """
        try:
            response = self.Connection.query('server.rehash', {'server': _serverorsid})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Connection.set_error(response)
                return None

            rehash = response['result']

            rehashObject = self.ModelRehash(
                        name=rehash['rehash_client']['name'] if 'rehash_client' in rehash and 'name' in rehash['rehash_client'] else None,
                        id=rehash['rehash_client']['id'] if 'rehash_client' in rehash and 'id' in rehash['rehash_client'] else None,
                        hostname=rehash['rehash_client']['hostname'] if 'rehash_client' in rehash and 'hostname' in rehash['rehash_client'] else None,
                        ip=rehash['rehash_client']['ip'] if 'rehash_client' in rehash and 'ip' in rehash['rehash_client'] else None,
                        server_port=rehash['rehash_client']['server_port'] if 'rehash_client' in rehash and 'server_port' in rehash['rehash_client'] else 0,
                        details=rehash['rehash_client']['details'] if 'rehash_client' in rehash and 'details' in rehash['rehash_client'] else None,
                        connected_since=rehash['rehash_client']['connected_since'] if 'rehash_client' in rehash and 'connected_since' in rehash['rehash_client'] else None,
                        idle_since=rehash['rehash_client']['idle_since'] if 'rehash_client' in rehash and 'idle_since' in rehash['rehash_client'] else None,
                        log=rehash['log'] if 'log' in rehash else [],
                        success=rehash['success'] if 'success' in rehash else False
                    )

            return rehashObject

        except TypeError as te:
            self.Logs.error(f'Type Error: {te}')
        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General Error: {err}')

    def connect(self, link: str) -> bool:
        """Make server link (connect) to another server.

        Right now you can only tell to link to servers from the directly connected server 
        (the one you are issuing JSON-RPC calls to), 
        you cannot (yet) tell remote server B to link to server C.

        Args:
            link (str): the server name to link to

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query('server.connect', {'link': link})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def disconnect(self, link: str) -> bool:
        """Terminate a server link (disconnect).

        This works for both directly connected servers and remote servers further up the network.

        Args:
            link (str): the server name to unlink from

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query('server.disconnect', {'link': link})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def module_list(self, _serverorsid: str = None) -> Union[list[ModelModules], None]:
        """Get the module list (list of loaded modules) on a server.

        IMPORTANT: This only works with remote servers if all servers on your network have rpc.modules.default.conf included.
        This is because then the JSON-RPC request/response is forwarded over the IRC network.

        Args:
            _serverorsid (str): The server name (or the SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API).. Defaults to None.

        Returns:
            ModelModules: if success you will find the object DB_MODULES
            bool: if False means that we have an error
        """
        try:
            response = self.Connection.query('server.module_list', {'server': _serverorsid})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Connection.set_error(response)
                return None

            modules = response['result']['list']

            for module in modules:
                self.DB_MODULES.append(
                    self.ModelModules(
                        name=module['name'] if 'name' in module else None,
                        version=module['version'] if 'version' in module else None,
                        author=module['author'] if 'author' in module else None,
                        description=module['description'] if 'description' in module else None,
                        third_party=module['third_party'] if 'third_party' in module else None,
                        permanent=module['permanent'] if 'permanent' in module else None,
                        permanent_but_reloadable=module['permanent_but_reloadable'] if 'permanent_but_reloadable' in module else None
                    )
                )

            return self.DB_MODULES

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')



import json.scanner
import json, socket, sys, os
import time, logging, random, asyncio
from typing import Literal, Union
from types import SimpleNamespace
from dataclasses import dataclass

class Live:

    @dataclass
    class ErrorModel:
        code: int
        message: str

    def __init__(self, path_to_socket_file: str, callback_object_instance: object, callback_method_name: str, debug_level: Literal[10, 20, 30, 40, 50] = 20) -> None:

        self.debug_level = debug_level
        self.Logs: logging
        self.__init_log_system()

        if not self.__check_unix_socket_file(path_to_socket_file=path_to_socket_file):
            self.Logs.critical(f'The socket file is not available, please check the full path of your socket file')
            self.Error.code = -1
            self.Error.message = 'The socket file is not available, please check the full path of your socket file'
            sys.exit('please check the full path of your socket file')

        self.to_run = getattr(callback_object_instance, callback_method_name)

        self.path_to_socket_file = path_to_socket_file

        self.request: str = ''
        self.str_response = ''
        self.json_response = ''

        self.connected: bool = True

        # Option 2 with Namespaces
        self.json_response_np: SimpleNamespace

        self.Error = self.ErrorModel(0, '')

    def __check_unix_socket_file(self, path_to_socket_file: str) -> bool:
        """Check provided full path to socket file if it exist

        Args:
            path_to_socket_file (str): Full path to unix socket file

        Returns:
            bool: True if path is correct else False
        """
        try:
            response = False

            if path_to_socket_file is None:
                return response
            
            if not os.path.exists(path_to_socket_file):
                return response

            response = True

            return response
        except NameError as nameerr:
            self.Logs.critical(f'NameError: {nameerr}')

    async def __send_to_permanent_unixsocket(self):
        try:

            sock = socket.socket(socket.AddressFamily.AF_UNIX, socket.SocketKind.SOCK_STREAM)
            sock.connect(self.path_to_socket_file)

            if not self.request:
                return None

            sock.sendall(f'{self.request}\r\n'.encode())

            while self.connected:
                # Recieve the data from the rpc server, decode it and split it
                response = sock.recv(4096).decode().split('\n')

                for bdata in response:
                    if bdata:
                        self.json_response = json.loads(bdata)
                        self.json_response_np: SimpleNamespace = json.loads(bdata, object_hook=lambda d: SimpleNamespace(**d))
                        self.to_run(self.json_response_np)

            sock.close()

        except AttributeError as attrerr:
            self.Logs.critical(f'AF_Unix Error: {attrerr}')
            sys.exit('AF_UNIX Are you sure you want to use Unix socket ?')
        except TimeoutError as timeouterr:
            self.Logs.critical(f'Timeout Error: {timeouterr}')
        except OSError as oserr:
            self.Logs.critical(f'System Error: {oserr}')
            sys.exit(3)
        except json.decoder.JSONDecodeError as jsondecoderror:
            self.Logs.critical(f'Json Decod Error: {jsondecoderror}')
        except Exception as err:
            self.Logs.error(f'General Error: {err}')

    def __init_log_system(self) -> None:
        """Init log system
        """
        # Init logs object
        self.Logs = logging
        self.Logs.basicConfig(level=self.debug_level,
                            encoding='UTF-8',
                            format='%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(funcName)s - %(message)s')

    def subscribe(self, param: dict = {"sources": ["all"]}):
        """Subscribe to the rpc server stream
        param exemple: 
        \n ["!debug","all"] would give you all log messages except for debug messages
        see: https://www.unrealircd.org/docs/List_of_all_log_messages
        Args:
            param (dict, optional): The ressources you want to subscribe. Defaults to {"sources": ["all"]}.
        """
        asyncio.run(self.query('log.subscribe', param=param))

    def unsubscribe(self):
        """Run a del timer to trigger an event and then unsubscribe from the stream
        """
        self.connected = False
        asyncio.run(self.query(method='rpc.del_timer', param={"timer_id":"timer_impossible_to_find_as_i_am_not_a_teapot"}))
        asyncio.run(self.query(method='log.unsubscribe'))

    async def query(self, method: Union[Literal['log.subscribe', 'log.unsubscribe'], str], param: dict = {}, id: int = 123, jsonrpc:str = '2.0') -> Union[str, any, None, bool]:
        """This method will use to run the queries

        Args:
            method (Union[Literal[&#39;stats.get&#39;, &#39;rpc.info&#39;,&#39;user.list&#39;], str]): The method to send to unrealircd
            param (dict, optional): the paramaters to send to unrealircd. Defaults to {}.
            id (int, optional): id of the request. Defaults to 123.
            jsonrpc (str, optional): jsonrpc. Defaults to '2.0'.

        Returns:
            str: The correct response
            None: no response from the server
            bool: False if there is an error occured
        """

        # data = '{"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 123}'
        get_method = method
        get_param = param

        if id == 123:
            rand = random.randint(1, 6000)
            get_id = int(time.time()) + rand

        else:
            get_id = id

        response = {
            "jsonrpc": jsonrpc,
            "method": get_method,
            "params": get_param,
            "id": get_id
        }

        self.request = json.dumps(response)

        await asyncio.gather(self.__send_to_permanent_unixsocket())

        if self.json_response == '':
            return False

        return self.json_response

    def set_error(self, json_error: dict):
        """set the error

        Args:
            code (int): the error code
            message (str): the message of the error
        """
        try:
            if 'error' in json_error:
                self.Error.code=json_error['error']['code']
                self.Error.message=json_error['error']['message']

        except KeyError as ke:
            self.Logs.error(ke)
        except Exception as err:
            self.Logs.error(err)



from types import SimpleNamespace
from typing import Union
from dataclasses import dataclass
from unrealircd_rpc_py.Connection import Connection

class Name_ban:

    @dataclass
    class ModelNameBan:
        type: str
        type_string: str
        set_by: str
        set_at: str
        expire_at: str
        set_at_string: str
        expire_at_string: str
        duration_string: str
        set_at_delta: int
        set_in_config: bool
        name: str
        reason: str

    DB_NAME_BANS: list[ModelNameBan] = []

    def __init__(self, Connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = Connection
        self.Logs = Connection.Logs
        self.Error = Connection.Error

    def list_(self) -> Union[list[ModelNameBan], None]:
        """List name bans (qlines).

        Returns:
            ModelNameBan: List of ModelNameBan, None if nothing see Error property
        """
        try:
            response = self.Connection.query(method='name_ban.list')

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            namebans = response['result']['list']

            for nameban in namebans:
                self.DB_NAME_BANS.append(
                        self.ModelNameBan(
                            type=nameban['type'] if 'type' in nameban else None,
                            type_string=nameban['type_string'] if 'type_string' in nameban else None,
                            set_by=nameban['set_by'] if 'set_by' in nameban else None,
                            set_at=nameban['set_at'] if 'set_at' in nameban else None,
                            expire_at=nameban['expier_at'] if 'expier_at' in nameban else None,
                            set_at_string=nameban['set_at_string'] if 'set_at_string' in nameban else None,
                            expire_at_string=nameban['expier_at_string'] if 'expier_at_string' in nameban else None,
                            duration_string=nameban['duration_string'] if 'duration_string' in nameban else None,
                            set_at_delta=nameban['set_at_delta'] if 'set_at_delta' in nameban else 0,
                            set_in_config=nameban['set_in_config'] if 'set_in_config' in nameban else False,
                            name=nameban['name'] if 'name' in nameban else None,
                            reason=nameban['reason'] if 'reason' in nameban else None
                            )
                )

            return self.DB_NAME_BANS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def get(self, name: str) -> Union[ModelNameBan, None]:
        """Retrieve all details of a single name ban (qline).

        Args:
            name (str): name of the ban, eg *C*h*a*n*s*e*r*v*

        Returns:
            Union[ModelNameBan, None, bool]: The Object ModelNameBan, None if nothing see Error property
        """
        try:
            response = self.Connection.query(method='name_ban.get', param={"name": name})

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                error = {"error": {"code": -1, "message": "Empty response"}}
                self.Connection.set_error(error)
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.set_error(response)
                return None

            nameban = response['result']['tkl']

            objectNameBan = self.ModelNameBan(
                            type=nameban['type'] if 'type' in nameban else None,
                            type_string=nameban['type_string'] if 'type_string' in nameban else None,
                            set_by=nameban['set_by'] if 'set_by' in nameban else None,
                            set_at=nameban['set_at'] if 'set_at' in nameban else None,
                            expire_at=nameban['expier_at'] if 'expier_at' in nameban else None,
                            set_at_string=nameban['set_at_string'] if 'set_at_string' in nameban else None,
                            expire_at_string=nameban['expier_at_string'] if 'expier_at_string' in nameban else None,
                            duration_string=nameban['duration_string'] if 'duration_string' in nameban else None,
                            set_at_delta=nameban['set_at_delta'] if 'set_at_delta' in nameban else 0,
                            set_in_config=nameban['set_in_config'] if 'set_in_config' in nameban else False,
                            name=nameban['name'] if 'name' in nameban else None,
                            reason=nameban['reason'] if 'reason' in nameban else None
                            )

            return objectNameBan

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add(self, name: str, reason: str, _set_by: str = None, _expire_at: str = None, _duration_string: str = None) -> bool:
        """Add a name ban (qline).

        Mandatory arguments (see structure of a name ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Name_ban#Structure_of_a_name_ban

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            reason (str): The reason of the ban
            _set_by (str, optional): Name of the person or server who set the ban. Default to None
            _expire_at (str, optional): Date/Time when the server ban will expire. NULL means: never. Default to None
            _duration_string (str, optional): How long the ban will last from this point in time (human printable). Uses "permanent" for forever. Default to None

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='name_ban.add', 
                param={"name": name, "reason": reason, "set_by": _set_by, "expire_at": _expire_at, "duration_string": _duration_string}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def del_(self, name: str, _set_by: str = None) -> bool:
        """Delete a name ban (LINE).

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            _set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            response = self.Connection.query(
                method='name_ban.del',
                param={"name": name, 'set_by': _set_by}
                )

            self.response_raw = response
            self.response_np = self.Connection.json_response_np

            if response is None:
                return False

            if 'error' in response:
                self.Connection.set_error(response)
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')
